from ._external import load, save, show, board_show
from .plugin_manager import PluginManager
from .base_plugin import BasePlugin


__all__ = ['load', 'save', 'show', 'board_show', 'BasePlugin', 'PluginManager']




from functools import partial
import warnings
from typing import Sequence
import random

from .common import GroupedColumns
from ..utils import LazyLoader

try:
    tf = LazyLoader('tf', globals(), 'tensorflow')
except (ImportError, ModuleNotFoundError):
    raise ImportError(
        'Could not import "tensorflow" library. Ensure library is '
        'installed correctly to use tensorflow dataloader functions') from None


def yield_data(columns, sample_names, shuffle=False):  # pragma: no cover
    if shuffle:
        sample_names = list(sample_names)
        random.shuffle(sample_names)
    for name in sample_names:
        yield tuple([col[name] for col in columns])


def make_tf_dataset(columns,
                    keys: Sequence[str] = None,
                    index_range: slice = None,
                    shuffle: bool = True):
    """
    Uses the hangar columns to make a tensorflow dataset. It uses
    `from_generator` function from `tensorflow.data.Dataset` with a generator
    function that wraps all the hangar columns. In such instances tensorflow
    Dataset does shuffle by loading the subset of data which can fit into the
    memory and shuffle that subset. Since it is not really a global shuffle
    `make_tf_dataset` accepts a `shuffle` argument which will be used by the
    generator to shuffle each time it is being called.

    .. warning::

        `tf.data.Dataset.from_generator` currently uses `tf.compat.v1.py_func()`
        internally. Hence the serialization function (`yield_data`) will not be
        serialized in a GraphDef. Therefore, you won't be able to serialize your
        model and restore it in a different environment if you use
        `make_tf_dataset`. The operation must run in the same address space as the
        Python program that calls tf.compat.v1.py_func(). If you are using
        distributed TensorFlow, you must run a tf.distribute.Server in the same
        process as the program that calls tf.compat.v1.py_func() and you must pin
        the created operation to a device in that server (e.g. using with
        tf.device():)

    Parameters
    ----------
    columns : :class:`~hangar.columns.column.Columns` or Sequence
        A column object, a tuple of column object or a list of column
        objects`
    keys : Sequence[str]
        An iterable of sample names. If given only those samples will fetched from
        the column
    index_range : slice
        A python slice object which will be used to find the subset of column.
        Argument `keys` takes priority over `index_range` i.e. if both are given,
        keys will be used and `index_range` will be ignored
    shuffle : bool
        generator uses this to decide a global shuffle accross all the samples is
        required or not. But user doesn't have any restriction on
        doing`column.shuffle()` on the returned column

    Examples
    --------
    >>> from hangar import Repository
    >>> from hangar import make_tf_dataset
    >>> import tensorflow as tf
    >>> tf.compat.v1.enable_eager_execution()
    >>> repo = Repository('.')
    >>> co = repo.checkout()
    >>> data = co.columns['mnist_data']
    >>> target = co.columns['mnist_target']
    >>> tf_dset = make_tf_dataset([data, target])
    >>> tf_dset = tf_dset.batch(512)
    >>> for bdata, btarget in tf_dset:
    ...     print(bdata.shape, btarget.shape)


    Returns
    -------
    :class:`tf.data.Dataset`
    """
    warnings.warn("Dataloaders are experimental in the current release.", UserWarning)
    gcolumns = GroupedColumns(columns, keys, index_range)
    generator = partial(yield_data, gcolumns.columns_col, gcolumns.sample_names, shuffle)
    res = tf.data.Dataset.from_generator(
        generator=generator,
        output_types=gcolumns.get_types(converter=tf.as_dtype),
        output_shapes=gcolumns.get_shapes(converter=tf.TensorShape))
    return res



"""Local LMDB Backend Implementation, Identifier: ``LMDB_30``

Backend Identifiers
===================

*  Backend: ``3``
*  Version: ``0``
*  Format Code: ``30``
*  Canonical Name: ``LMDB_30``

Storage Method
==============

*  This module is meant to handle string typed data which is of any size. IO
   is performed via the LMDB storage system.

*  This module does not compress values upon writing, the full (uncompressed)
   value of the text is written to the DB for each key.

*  For each LMDB file generated, data is indexed by keys which are generated
   in lexicographically sorted order of key length 4. Keys consist of 4 characters
   chosen from an alphabet consisting of ASCII digits, lowercase letters, and
   upercase letters. Within a single write instance (when an LMDB file is created
   and written to), lexicographically sorted permutations of the chosen characters
   are used as key indexes.

   This means that for each LMDB file written in a repo, the sequence of generated
   index keys will be identical, even though two databases with the same key will
   store different values. As such, the File UID is crucial in order to identify
   a unique db/index key combo to access a particular value by.

*  There is no limit to the size which each record can occupy. Data is stored
   "as-is" and is uncompressed. Reading the data back will return the exact
   data stored (regardless of how large the data record is).

*  On read and write of all samples the xxhash64_hexdigest is calculated for
   the raw data bytes. This is to ensure that all data in == data out of the
   lmdb files. That way even if a file is manually edited we have a quick way
   to tell that things are not as they should be. (full data hash digests may
   not be calculated every time a read is performed).

Compression Options
===================

None

Record Format
=============

Fields Recorded for Each Array
------------------------------

*  Format Code
*  File UID
*  Row Index

Examples
--------

1)  Adding the first piece of data to a file:

    *  File UID: "rlUK3C"
    *  Row Index: "0123"
    *  xxhash64_hexdigest: 8067007c0f05c359

    ``Record Data => "30:rlUK3C:0123:8067007c0f05c359"``

2)  Adding a second piece of data:

    *  File UID: "rlUK3C"
    *  Row Index: "0124"
    *  xxhash64_hexdigest: b89f873d3d153a9c

    ``Record Data => "30:rlUK3C:0124:b89f873d3d153a9c"``

3)  Adding a the 500th piece of data:

    *  File UID: "rlUK3C"
    *  Row Index: "01AU"
    *  xxhash64_hexdigest: cf3fc53cad153a5a

    ``Record Data => "30:rlUK3C:01AU:cf3fc53cad153a5a"``
"""
import os
import shutil
import string
from collections import ChainMap
from contextlib import suppress
from functools import partial
from itertools import permutations
from pathlib import Path
from typing import Optional

import lmdb
from xxhash import xxh64_hexdigest

from .specs import LMDB_30_DataHashSpec
from ..constants import DIR_DATA_REMOTE, DIR_DATA_STAGE, DIR_DATA_STORE, DIR_DATA
from ..op_state import reader_checkout_only, writer_checkout_only
from ..utils import random_string
from ..typesystem import Descriptor, OneOf, EmptyDict, checkedmeta


LMDB_SETTINGS = {
    'map_size': 300_000_000,
    'meminit': False,
    'subdir': True,
    'lock': True,
    'max_spare_txns': 4,
}
_FmtCode = '30'


def _lexicographic_keys():
    lexicographic_ids = ''.join([
        string.digits,
        string.ascii_uppercase,
        string.ascii_lowercase,
    ])
    # permutations generates results in lexicographic order
    # total of 14_776_336 total ids can be generated with
    # a row_id consisting of 4 characters. This is more keys than
    # we will ever allow in a single LMDB database
    p = permutations(lexicographic_ids, 4)

    for perm in p:
        res = ''.join(perm)
        yield res


def lmdb_30_encode(uid: str, row_idx: int, checksum: str) -> bytes:
    res = f'30:{uid}:{row_idx}:{checksum}'
    return res.encode()


@OneOf(['<class\'str\'>', str])
class AllowedDtypes(Descriptor):
    pass


class LMDB_30_Options(metaclass=checkedmeta):
    _dtype = AllowedDtypes()
    _backend_options = EmptyDict()

    def __init__(self, backend_options, dtype, *args, **kwargs):
        if backend_options is None:
            backend_options = self.default_options
        self._backend_options = backend_options
        self._dtype = dtype

    @property
    def default_options(self):
        return {}

    @property
    def backend_options(self):
        return self._backend_options

    @property
    def init_requires(self):
        return ('repo_path',)


class LMDB_30_FileHandles:

    def __init__(self, repo_path: Path, *args, **kwargs):

        self.path: Path = repo_path

        self.rFp = {}
        self.wFp = {}
        self.Fp = ChainMap(self.rFp, self.wFp)

        self.mode: Optional[str] = None
        self.w_uid: Optional[str] = None
        self.row_idx: Optional[str] = None
        self._dflt_backend_opts: Optional[dict] = None

        self.STAGEDIR: Path = Path(self.path, DIR_DATA_STAGE, _FmtCode)
        self.REMOTEDIR: Path = Path(self.path, DIR_DATA_REMOTE, _FmtCode)
        self.STOREDIR: Path = Path(self.path, DIR_DATA_STORE, _FmtCode)
        self.DATADIR: Path = Path(self.path, DIR_DATA, _FmtCode)
        self.DATADIR.mkdir(exist_ok=True)

    def __enter__(self):

        return self

    def __exit__(self, *exc):
        return

    @reader_checkout_only
    def __getstate__(self) -> dict:
        """ensure multiprocess operations can pickle relevant data.
        """
        self.close()
        state = self.__dict__.copy()
        del state['rFp']
        del state['wFp']
        del state['Fp']
        return state

    def __setstate__(self, state: dict) -> None:  # pragma: no cover
        """ensure multiprocess operations can pickle relevant data.
        """
        self.__dict__.update(state)
        self.rFp = {}
        self.wFp = {}
        self.Fp = ChainMap(self.rFp, self.wFp)

    @property
    def backend_opts(self):
        return self._dflt_backend_opts

    @writer_checkout_only
    def _backend_opts_set(self, val):
        """Nonstandard descriptor method. See notes in ``backend_opts.setter``.
        """
        self._dflt_backend_opts = val
        return

    @backend_opts.setter
    def backend_opts(self, value):
        """
        Using seperate setter method (with ``@writer_checkout_only`` decorator
        applied) due to bug in python <3.8.

        From: https://bugs.python.org/issue19072
            > The classmethod decorator when applied to a function of a class,
            > does not honour the descriptor binding protocol for whatever it
            > wraps. This means it will fail when applied around a function which
            > has a decorator already applied to it and where that decorator
            > expects that the descriptor binding protocol is executed in order
            > to properly bind the function to the class.
        """
        return self._backend_opts_set(value)

    def open(self, mode: str, *, remote_operation: bool = False):
        """Open an lmdb file handle.

        Parameters
        ----------
        mode : str
            one of `r` or `a` for read only / read-write.
        remote_operation : optional, kwarg only, bool
            if this lmdb data is being created from a remote fetch operation, then
            we don't open any files for reading, and only open files for writing
            which exist in the remote data dir. (default is false, which means that
            write operations use the stage data dir and read operations use data store
            dir)
        """
        self.mode = mode
        if self.mode == 'a':
            process_dir = self.REMOTEDIR if remote_operation else self.STAGEDIR
            process_dir.mkdir(exist_ok=True)
            for uidpth in process_dir.iterdir():
                if uidpth.suffix == '.lmdbdir':
                    file_pth = self.DATADIR.joinpath(uidpth.stem)
                    self.rFp[uidpth.stem] = partial(lmdb.open, str(file_pth), readonly=True,
                                                    **LMDB_SETTINGS)

        if not remote_operation:
            if not self.STOREDIR.is_dir():
                return
            for uidpth in self.STOREDIR.iterdir():
                if uidpth.suffix == '.lmdbdir':
                    file_pth = self.DATADIR.joinpath(uidpth.stem)
                    self.rFp[uidpth.stem] = partial(lmdb.open, str(file_pth), readonly=True,
                                                    **LMDB_SETTINGS)

    def close(self):
        """Close a file handle after writes have been completed

        behavior changes depending on write-enable or read-only file

        Returns
        -------
        bool
            True if success, otherwise False.
        """
        if self.mode == 'a':
            for uid in list(self.wFp.keys()):
                with suppress(AttributeError):
                    self.wFp[uid].close()
                del self.wFp[uid]
            self.w_uid = None
            self.row_idx = None

        for uid in list(self.rFp.keys()):
            with suppress(AttributeError):
                self.rFp[uid].close()
            del self.rFp[uid]

    @staticmethod
    def delete_in_process_data(repo_path: Path, *, remote_operation=False) -> None:
        """Removes some set of files entirely from the stage/remote directory.

        DANGER ZONE. This should essentially only be used to perform hard resets
        of the repository state.

        Parameters
        ----------
        repo_path : Path
            path to the repository on disk
        remote_operation : optional, kwarg only, bool
            If true, modify contents of the remote_dir, if false (default) modify
            contents of the staging directory.
        """
        data_dir = Path(repo_path, DIR_DATA, _FmtCode)
        PDIR = DIR_DATA_STAGE if not remote_operation else DIR_DATA_REMOTE
        process_dir = Path(repo_path, PDIR, _FmtCode)
        if not process_dir.is_dir():
            return

        for uidpth in process_dir.iterdir():
            if uidpth.suffix == '.lmdbdir':
                os.remove(process_dir.joinpath(uidpth.name))
                db_dir = data_dir.joinpath(uidpth.stem)
                shutil.rmtree(str(db_dir))
        os.rmdir(process_dir)

    def _create_schema(self, *, remote_operation: bool = False):

        uid = random_string()
        db_dir_path = self.DATADIR.joinpath(f'{uid}')
        self.wFp[uid] = lmdb.open(str(db_dir_path), **LMDB_SETTINGS)

        self.w_uid = uid
        self.row_idx = _lexicographic_keys()

        process_dir = self.REMOTEDIR if remote_operation else self.STAGEDIR
        Path(process_dir, f'{uid}.lmdbdir').touch()

    def read_data(self, hashVal: LMDB_30_DataHashSpec) -> str:
        """Read data from an hdf5 file handle at the specified locations

        Parameters
        ----------
        hashVal : LMDB_30_DataHashSpec
            record specification parsed from its serialized store val in lmdb.

        Returns
        -------
        str
            requested data.
        """
        try:
            with self.Fp[hashVal.uid].begin(write=False) as txn:
                res = txn.get(hashVal.row_idx.encode(), default=False)
                if res is False:
                    raise RuntimeError(hashVal)
        except AttributeError:
            self.Fp[hashVal.uid] = self.Fp[hashVal.uid]()
            return self.read_data(hashVal)
        except KeyError:
            process_dir = self.STAGEDIR if self.mode == 'a' else self.STOREDIR
            if Path(process_dir, f'{hashVal.uid}.lmdbdir').is_file():
                file_pth = self.DATADIR.joinpath(hashVal.uid)
                self.rFp[hashVal.uid] = lmdb.open(str(file_pth), readonly=True, **LMDB_SETTINGS)
                return self.read_data(hashVal)
            else:
                raise

        out = res.decode()
        if xxh64_hexdigest(res) != hashVal.checksum:
            raise RuntimeError(
                f'DATA CORRUPTION Checksum {xxh64_hexdigest(res)} != recorded {hashVal}')
        return out

    def write_data(self, data: str, *, remote_operation: bool = False) -> bytes:
        """verifies correctness of array data and performs write operation.

        Parameters
        ----------
        data: str
            data to write to group.
        remote_operation : optional, kwarg only, bool
            If this is a remote process which is adding data, any necessary
            hdf5 dataset files will be created in the remote data dir instead
            of the stage directory. (default is False, which is for a regular
            access process)

        Returns
        -------
        bytes
            string identifying the collection dataset and collection dim-0 index
            which the array can be accessed at.
        """
        encoded_data = data.encode()
        checksum = xxh64_hexdigest(encoded_data)

        if self.w_uid in self.wFp:
            try:
                row_idx = next(self.row_idx)
            except StopIteration:
                self._create_schema(remote_operation=remote_operation)
                return self.write_data(data, remote_operation=remote_operation)
        else:
            self._create_schema(remote_operation=remote_operation)
            return self.write_data(data, remote_operation=remote_operation)

        encoded_row_idx = row_idx.encode()
        try:
            with self.wFp[self.w_uid].begin(write=True) as txn:
                txn.put(encoded_row_idx, encoded_data, append=True)
        except lmdb.MapFullError:
            self._create_schema(remote_operation=remote_operation)
            return self.write_data(data, remote_operation=remote_operation)

        return lmdb_30_encode(self.w_uid, row_idx, checksum)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hangar_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hangar_service.proto',
  package='hangar',
  syntax='proto3',
  serialized_options=_b('H\001'),
  serialized_pb=_b('\n\x14hangar_service.proto\x12\x06hangar\"+\n\nErrorProto\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x03\x12\x0f\n\x07message\x18\x02 \x01(\t\",\n\x0c\x42ranchRecord\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06\x63ommit\x18\x02 \x01(\t\"*\n\nHashRecord\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0e\n\x06\x64igest\x18\x02 \x01(\t\"9\n\x0c\x43ommitRecord\x12\x0e\n\x06parent\x18\x01 \x01(\x0c\x12\x0b\n\x03ref\x18\x02 \x01(\x0c\x12\x0c\n\x04spec\x18\x03 \x01(\x0c\",\n\x0cSchemaRecord\x12\x0e\n\x06\x64igest\x18\x01 \x01(\t\x12\x0c\n\x04\x62lob\x18\x02 \x01(\x0c\"\r\n\x0bPingRequest\"\x1b\n\tPingReply\x12\x0e\n\x06result\x18\x01 \x01(\t\"\x18\n\x16GetClientConfigRequest\"\xa2\x01\n\x14GetClientConfigReply\x12\x38\n\x06\x63onfig\x18\x01 \x03(\x0b\x32(.hangar.GetClientConfigReply.ConfigEntry\x12!\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x12.hangar.ErrorProto\x1a-\n\x0b\x43onfigEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"=\n\x18\x46\x65tchBranchRecordRequest\x12!\n\x03rec\x18\x01 \x01(\x0b\x32\x14.hangar.BranchRecord\"^\n\x16\x46\x65tchBranchRecordReply\x12!\n\x03rec\x18\x01 \x01(\x0b\x32\x14.hangar.BranchRecord\x12!\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x12.hangar.ErrorProto\"s\n\x10\x46\x65tchDataRequest\x12\x10\n\x08raw_data\x18\x01 \x01(\x0c\x12\x13\n\x0b\x63omp_nbytes\x18\x02 \x01(\x03\x12\x15\n\runcomp_nbytes\x18\x03 \x01(\x03\x12!\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x12.hangar.ErrorProto\"q\n\x0e\x46\x65tchDataReply\x12\x10\n\x08raw_data\x18\x01 \x01(\x0c\x12\x13\n\x0b\x63omp_nbytes\x18\x02 \x01(\x03\x12\x15\n\runcomp_nbytes\x18\x03 \x01(\x03\x12!\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x12.hangar.ErrorProto\"$\n\x12\x46\x65tchCommitRequest\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\"\x84\x01\n\x10\x46\x65tchCommitReply\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\x17\n\x0ftotal_byte_size\x18\x02 \x01(\x03\x12$\n\x06record\x18\x03 \x01(\x0b\x32\x14.hangar.CommitRecord\x12!\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x12.hangar.ErrorProto\"7\n\x12\x46\x65tchSchemaRequest\x12!\n\x03rec\x18\x01 \x01(\x0b\x32\x14.hangar.SchemaRecord\"X\n\x10\x46\x65tchSchemaReply\x12!\n\x03rec\x18\x01 \x01(\x0b\x32\x14.hangar.SchemaRecord\x12!\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x12.hangar.ErrorProto\"<\n\x17PushBranchRecordRequest\x12!\n\x03rec\x18\x01 \x01(\x0b\x32\x14.hangar.BranchRecord\":\n\x15PushBranchRecordReply\x12!\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x12.hangar.ErrorProto\"r\n\x0fPushDataRequest\x12\x10\n\x08raw_data\x18\x01 \x01(\x0c\x12\x13\n\x0b\x63omp_nbytes\x18\x02 \x01(\x03\x12\x15\n\runcomp_nbytes\x18\x03 \x01(\x03\x12!\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x12.hangar.ErrorProto\"2\n\rPushDataReply\x12!\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x12.hangar.ErrorProto\"b\n\x11PushCommitRequest\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\x17\n\x0ftotal_byte_size\x18\x02 \x01(\x03\x12$\n\x06record\x18\x03 \x01(\x0b\x32\x14.hangar.CommitRecord\"4\n\x0fPushCommitReply\x12!\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x12.hangar.ErrorProto\"6\n\x11PushSchemaRequest\x12!\n\x03rec\x18\x01 \x01(\x0b\x32\x14.hangar.SchemaRecord\"4\n\x0fPushSchemaReply\x12!\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x12.hangar.ErrorProto\"R\n\x19\x46indMissingCommitsRequest\x12\x0f\n\x07\x63ommits\x18\x01 \x03(\t\x12$\n\x06\x62ranch\x18\x02 \x01(\x0b\x32\x14.hangar.BranchRecord\"s\n\x17\x46indMissingCommitsReply\x12\x0f\n\x07\x63ommits\x18\x01 \x03(\t\x12$\n\x06\x62ranch\x18\x02 \x01(\x0b\x32\x14.hangar.BranchRecord\x12!\n\x05\x65rror\x18\x03 \x01(\x0b\x32\x12.hangar.ErrorProto\"W\n\x1d\x46indMissingHashRecordsRequest\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\r\n\x05hashs\x18\x02 \x01(\x0c\x12\x17\n\x0ftotal_byte_size\x18\x03 \x01(\x03\"x\n\x1b\x46indMissingHashRecordsReply\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\r\n\x05hashs\x18\x02 \x01(\x0c\x12\x17\n\x0ftotal_byte_size\x18\x03 \x01(\x03\x12!\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x12.hangar.ErrorProto\"C\n\x19\x46indMissingSchemasRequest\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\x16\n\x0eschema_digests\x18\x02 \x03(\t\"d\n\x17\x46indMissingSchemasReply\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\x16\n\x0eschema_digests\x18\x02 \x03(\t\x12!\n\x05\x65rror\x18\x03 \x01(\x0b\x32\x12.hangar.ErrorProto2\xc5\n\n\rHangarService\x12\x30\n\x04PING\x12\x13.hangar.PingRequest\x1a\x11.hangar.PingReply\"\x00\x12Q\n\x0fGetClientConfig\x12\x1e.hangar.GetClientConfigRequest\x1a\x1c.hangar.GetClientConfigReply\"\x00\x12W\n\x11\x46\x65tchBranchRecord\x12 .hangar.FetchBranchRecordRequest\x1a\x1e.hangar.FetchBranchRecordReply\"\x00\x12\x43\n\tFetchData\x12\x18.hangar.FetchDataRequest\x1a\x16.hangar.FetchDataReply\"\x00(\x01\x30\x01\x12G\n\x0b\x46\x65tchCommit\x12\x1a.hangar.FetchCommitRequest\x1a\x18.hangar.FetchCommitReply\"\x00\x30\x01\x12\x45\n\x0b\x46\x65tchSchema\x12\x1a.hangar.FetchSchemaRequest\x1a\x18.hangar.FetchSchemaReply\"\x00\x12T\n\x10PushBranchRecord\x12\x1f.hangar.PushBranchRecordRequest\x1a\x1d.hangar.PushBranchRecordReply\"\x00\x12>\n\x08PushData\x12\x17.hangar.PushDataRequest\x1a\x15.hangar.PushDataReply\"\x00(\x01\x12\x44\n\nPushCommit\x12\x19.hangar.PushCommitRequest\x1a\x17.hangar.PushCommitReply\"\x00(\x01\x12\x42\n\nPushSchema\x12\x19.hangar.PushSchemaRequest\x1a\x17.hangar.PushSchemaReply\"\x00\x12_\n\x17\x46\x65tchFindMissingCommits\x12!.hangar.FindMissingCommitsRequest\x1a\x1f.hangar.FindMissingCommitsReply\"\x00\x12o\n\x1b\x46\x65tchFindMissingHashRecords\x12%.hangar.FindMissingHashRecordsRequest\x1a#.hangar.FindMissingHashRecordsReply\"\x00(\x01\x30\x01\x12_\n\x17\x46\x65tchFindMissingSchemas\x12!.hangar.FindMissingSchemasRequest\x1a\x1f.hangar.FindMissingSchemasReply\"\x00\x12^\n\x16PushFindMissingCommits\x12!.hangar.FindMissingCommitsRequest\x1a\x1f.hangar.FindMissingCommitsReply\"\x00\x12n\n\x1aPushFindMissingHashRecords\x12%.hangar.FindMissingHashRecordsRequest\x1a#.hangar.FindMissingHashRecordsReply\"\x00(\x01\x30\x01\x12^\n\x16PushFindMissingSchemas\x12!.hangar.FindMissingSchemasRequest\x1a\x1f.hangar.FindMissingSchemasReply\"\x00\x42\x02H\x01\x62\x06proto3')
)




_ERRORPROTO = _descriptor.Descriptor(
  name='ErrorProto',
  full_name='hangar.ErrorProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='hangar.ErrorProto.code', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='hangar.ErrorProto.message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=32,
  serialized_end=75,
)


_BRANCHRECORD = _descriptor.Descriptor(
  name='BranchRecord',
  full_name='hangar.BranchRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='hangar.BranchRecord.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.BranchRecord.commit', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=77,
  serialized_end=121,
)


_HASHRECORD = _descriptor.Descriptor(
  name='HashRecord',
  full_name='hangar.HashRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='hangar.HashRecord.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='digest', full_name='hangar.HashRecord.digest', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=123,
  serialized_end=165,
)


_COMMITRECORD = _descriptor.Descriptor(
  name='CommitRecord',
  full_name='hangar.CommitRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='hangar.CommitRecord.parent', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ref', full_name='hangar.CommitRecord.ref', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='spec', full_name='hangar.CommitRecord.spec', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=167,
  serialized_end=224,
)


_SCHEMARECORD = _descriptor.Descriptor(
  name='SchemaRecord',
  full_name='hangar.SchemaRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='digest', full_name='hangar.SchemaRecord.digest', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blob', full_name='hangar.SchemaRecord.blob', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=226,
  serialized_end=270,
)


_PINGREQUEST = _descriptor.Descriptor(
  name='PingRequest',
  full_name='hangar.PingRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=272,
  serialized_end=285,
)


_PINGREPLY = _descriptor.Descriptor(
  name='PingReply',
  full_name='hangar.PingReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='hangar.PingReply.result', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=287,
  serialized_end=314,
)


_GETCLIENTCONFIGREQUEST = _descriptor.Descriptor(
  name='GetClientConfigRequest',
  full_name='hangar.GetClientConfigRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=316,
  serialized_end=340,
)


_GETCLIENTCONFIGREPLY_CONFIGENTRY = _descriptor.Descriptor(
  name='ConfigEntry',
  full_name='hangar.GetClientConfigReply.ConfigEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='hangar.GetClientConfigReply.ConfigEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='hangar.GetClientConfigReply.ConfigEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=460,
  serialized_end=505,
)

_GETCLIENTCONFIGREPLY = _descriptor.Descriptor(
  name='GetClientConfigReply',
  full_name='hangar.GetClientConfigReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='config', full_name='hangar.GetClientConfigReply.config', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.GetClientConfigReply.error', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_GETCLIENTCONFIGREPLY_CONFIGENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=343,
  serialized_end=505,
)


_FETCHBRANCHRECORDREQUEST = _descriptor.Descriptor(
  name='FetchBranchRecordRequest',
  full_name='hangar.FetchBranchRecordRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rec', full_name='hangar.FetchBranchRecordRequest.rec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=507,
  serialized_end=568,
)


_FETCHBRANCHRECORDREPLY = _descriptor.Descriptor(
  name='FetchBranchRecordReply',
  full_name='hangar.FetchBranchRecordReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rec', full_name='hangar.FetchBranchRecordReply.rec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FetchBranchRecordReply.error', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=570,
  serialized_end=664,
)


_FETCHDATAREQUEST = _descriptor.Descriptor(
  name='FetchDataRequest',
  full_name='hangar.FetchDataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='raw_data', full_name='hangar.FetchDataRequest.raw_data', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='comp_nbytes', full_name='hangar.FetchDataRequest.comp_nbytes', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uncomp_nbytes', full_name='hangar.FetchDataRequest.uncomp_nbytes', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FetchDataRequest.error', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=666,
  serialized_end=781,
)


_FETCHDATAREPLY = _descriptor.Descriptor(
  name='FetchDataReply',
  full_name='hangar.FetchDataReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='raw_data', full_name='hangar.FetchDataReply.raw_data', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='comp_nbytes', full_name='hangar.FetchDataReply.comp_nbytes', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uncomp_nbytes', full_name='hangar.FetchDataReply.uncomp_nbytes', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FetchDataReply.error', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=783,
  serialized_end=896,
)


_FETCHCOMMITREQUEST = _descriptor.Descriptor(
  name='FetchCommitRequest',
  full_name='hangar.FetchCommitRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.FetchCommitRequest.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=898,
  serialized_end=934,
)


_FETCHCOMMITREPLY = _descriptor.Descriptor(
  name='FetchCommitReply',
  full_name='hangar.FetchCommitReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.FetchCommitReply.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_byte_size', full_name='hangar.FetchCommitReply.total_byte_size', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='record', full_name='hangar.FetchCommitReply.record', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FetchCommitReply.error', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=937,
  serialized_end=1069,
)


_FETCHSCHEMAREQUEST = _descriptor.Descriptor(
  name='FetchSchemaRequest',
  full_name='hangar.FetchSchemaRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rec', full_name='hangar.FetchSchemaRequest.rec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1071,
  serialized_end=1126,
)


_FETCHSCHEMAREPLY = _descriptor.Descriptor(
  name='FetchSchemaReply',
  full_name='hangar.FetchSchemaReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rec', full_name='hangar.FetchSchemaReply.rec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FetchSchemaReply.error', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1128,
  serialized_end=1216,
)


_PUSHBRANCHRECORDREQUEST = _descriptor.Descriptor(
  name='PushBranchRecordRequest',
  full_name='hangar.PushBranchRecordRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rec', full_name='hangar.PushBranchRecordRequest.rec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1218,
  serialized_end=1278,
)


_PUSHBRANCHRECORDREPLY = _descriptor.Descriptor(
  name='PushBranchRecordReply',
  full_name='hangar.PushBranchRecordReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.PushBranchRecordReply.error', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1280,
  serialized_end=1338,
)


_PUSHDATAREQUEST = _descriptor.Descriptor(
  name='PushDataRequest',
  full_name='hangar.PushDataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='raw_data', full_name='hangar.PushDataRequest.raw_data', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='comp_nbytes', full_name='hangar.PushDataRequest.comp_nbytes', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uncomp_nbytes', full_name='hangar.PushDataRequest.uncomp_nbytes', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.PushDataRequest.error', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1340,
  serialized_end=1454,
)


_PUSHDATAREPLY = _descriptor.Descriptor(
  name='PushDataReply',
  full_name='hangar.PushDataReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.PushDataReply.error', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1456,
  serialized_end=1506,
)


_PUSHCOMMITREQUEST = _descriptor.Descriptor(
  name='PushCommitRequest',
  full_name='hangar.PushCommitRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.PushCommitRequest.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_byte_size', full_name='hangar.PushCommitRequest.total_byte_size', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='record', full_name='hangar.PushCommitRequest.record', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1508,
  serialized_end=1606,
)


_PUSHCOMMITREPLY = _descriptor.Descriptor(
  name='PushCommitReply',
  full_name='hangar.PushCommitReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.PushCommitReply.error', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1608,
  serialized_end=1660,
)


_PUSHSCHEMAREQUEST = _descriptor.Descriptor(
  name='PushSchemaRequest',
  full_name='hangar.PushSchemaRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rec', full_name='hangar.PushSchemaRequest.rec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1662,
  serialized_end=1716,
)


_PUSHSCHEMAREPLY = _descriptor.Descriptor(
  name='PushSchemaReply',
  full_name='hangar.PushSchemaReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.PushSchemaReply.error', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1718,
  serialized_end=1770,
)


_FINDMISSINGCOMMITSREQUEST = _descriptor.Descriptor(
  name='FindMissingCommitsRequest',
  full_name='hangar.FindMissingCommitsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commits', full_name='hangar.FindMissingCommitsRequest.commits', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='branch', full_name='hangar.FindMissingCommitsRequest.branch', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1772,
  serialized_end=1854,
)


_FINDMISSINGCOMMITSREPLY = _descriptor.Descriptor(
  name='FindMissingCommitsReply',
  full_name='hangar.FindMissingCommitsReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commits', full_name='hangar.FindMissingCommitsReply.commits', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='branch', full_name='hangar.FindMissingCommitsReply.branch', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FindMissingCommitsReply.error', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1856,
  serialized_end=1971,
)


_FINDMISSINGHASHRECORDSREQUEST = _descriptor.Descriptor(
  name='FindMissingHashRecordsRequest',
  full_name='hangar.FindMissingHashRecordsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.FindMissingHashRecordsRequest.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hashs', full_name='hangar.FindMissingHashRecordsRequest.hashs', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_byte_size', full_name='hangar.FindMissingHashRecordsRequest.total_byte_size', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1973,
  serialized_end=2060,
)


_FINDMISSINGHASHRECORDSREPLY = _descriptor.Descriptor(
  name='FindMissingHashRecordsReply',
  full_name='hangar.FindMissingHashRecordsReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.FindMissingHashRecordsReply.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hashs', full_name='hangar.FindMissingHashRecordsReply.hashs', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_byte_size', full_name='hangar.FindMissingHashRecordsReply.total_byte_size', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FindMissingHashRecordsReply.error', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2062,
  serialized_end=2182,
)


_FINDMISSINGSCHEMASREQUEST = _descriptor.Descriptor(
  name='FindMissingSchemasRequest',
  full_name='hangar.FindMissingSchemasRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.FindMissingSchemasRequest.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='schema_digests', full_name='hangar.FindMissingSchemasRequest.schema_digests', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2184,
  serialized_end=2251,
)


_FINDMISSINGSCHEMASREPLY = _descriptor.Descriptor(
  name='FindMissingSchemasReply',
  full_name='hangar.FindMissingSchemasReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.FindMissingSchemasReply.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='schema_digests', full_name='hangar.FindMissingSchemasReply.schema_digests', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FindMissingSchemasReply.error', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2253,
  serialized_end=2353,
)

_GETCLIENTCONFIGREPLY_CONFIGENTRY.containing_type = _GETCLIENTCONFIGREPLY
_GETCLIENTCONFIGREPLY.fields_by_name['config'].message_type = _GETCLIENTCONFIGREPLY_CONFIGENTRY
_GETCLIENTCONFIGREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FETCHBRANCHRECORDREQUEST.fields_by_name['rec'].message_type = _BRANCHRECORD
_FETCHBRANCHRECORDREPLY.fields_by_name['rec'].message_type = _BRANCHRECORD
_FETCHBRANCHRECORDREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FETCHDATAREQUEST.fields_by_name['error'].message_type = _ERRORPROTO
_FETCHDATAREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FETCHCOMMITREPLY.fields_by_name['record'].message_type = _COMMITRECORD
_FETCHCOMMITREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FETCHSCHEMAREQUEST.fields_by_name['rec'].message_type = _SCHEMARECORD
_FETCHSCHEMAREPLY.fields_by_name['rec'].message_type = _SCHEMARECORD
_FETCHSCHEMAREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_PUSHBRANCHRECORDREQUEST.fields_by_name['rec'].message_type = _BRANCHRECORD
_PUSHBRANCHRECORDREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_PUSHDATAREQUEST.fields_by_name['error'].message_type = _ERRORPROTO
_PUSHDATAREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_PUSHCOMMITREQUEST.fields_by_name['record'].message_type = _COMMITRECORD
_PUSHCOMMITREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_PUSHSCHEMAREQUEST.fields_by_name['rec'].message_type = _SCHEMARECORD
_PUSHSCHEMAREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FINDMISSINGCOMMITSREQUEST.fields_by_name['branch'].message_type = _BRANCHRECORD
_FINDMISSINGCOMMITSREPLY.fields_by_name['branch'].message_type = _BRANCHRECORD
_FINDMISSINGCOMMITSREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FINDMISSINGHASHRECORDSREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FINDMISSINGSCHEMASREPLY.fields_by_name['error'].message_type = _ERRORPROTO
DESCRIPTOR.message_types_by_name['ErrorProto'] = _ERRORPROTO
DESCRIPTOR.message_types_by_name['BranchRecord'] = _BRANCHRECORD
DESCRIPTOR.message_types_by_name['HashRecord'] = _HASHRECORD
DESCRIPTOR.message_types_by_name['CommitRecord'] = _COMMITRECORD
DESCRIPTOR.message_types_by_name['SchemaRecord'] = _SCHEMARECORD
DESCRIPTOR.message_types_by_name['PingRequest'] = _PINGREQUEST
DESCRIPTOR.message_types_by_name['PingReply'] = _PINGREPLY
DESCRIPTOR.message_types_by_name['GetClientConfigRequest'] = _GETCLIENTCONFIGREQUEST
DESCRIPTOR.message_types_by_name['GetClientConfigReply'] = _GETCLIENTCONFIGREPLY
DESCRIPTOR.message_types_by_name['FetchBranchRecordRequest'] = _FETCHBRANCHRECORDREQUEST
DESCRIPTOR.message_types_by_name['FetchBranchRecordReply'] = _FETCHBRANCHRECORDREPLY
DESCRIPTOR.message_types_by_name['FetchDataRequest'] = _FETCHDATAREQUEST
DESCRIPTOR.message_types_by_name['FetchDataReply'] = _FETCHDATAREPLY
DESCRIPTOR.message_types_by_name['FetchCommitRequest'] = _FETCHCOMMITREQUEST
DESCRIPTOR.message_types_by_name['FetchCommitReply'] = _FETCHCOMMITREPLY
DESCRIPTOR.message_types_by_name['FetchSchemaRequest'] = _FETCHSCHEMAREQUEST
DESCRIPTOR.message_types_by_name['FetchSchemaReply'] = _FETCHSCHEMAREPLY
DESCRIPTOR.message_types_by_name['PushBranchRecordRequest'] = _PUSHBRANCHRECORDREQUEST
DESCRIPTOR.message_types_by_name['PushBranchRecordReply'] = _PUSHBRANCHRECORDREPLY
DESCRIPTOR.message_types_by_name['PushDataRequest'] = _PUSHDATAREQUEST
DESCRIPTOR.message_types_by_name['PushDataReply'] = _PUSHDATAREPLY
DESCRIPTOR.message_types_by_name['PushCommitRequest'] = _PUSHCOMMITREQUEST
DESCRIPTOR.message_types_by_name['PushCommitReply'] = _PUSHCOMMITREPLY
DESCRIPTOR.message_types_by_name['PushSchemaRequest'] = _PUSHSCHEMAREQUEST
DESCRIPTOR.message_types_by_name['PushSchemaReply'] = _PUSHSCHEMAREPLY
DESCRIPTOR.message_types_by_name['FindMissingCommitsRequest'] = _FINDMISSINGCOMMITSREQUEST
DESCRIPTOR.message_types_by_name['FindMissingCommitsReply'] = _FINDMISSINGCOMMITSREPLY
DESCRIPTOR.message_types_by_name['FindMissingHashRecordsRequest'] = _FINDMISSINGHASHRECORDSREQUEST
DESCRIPTOR.message_types_by_name['FindMissingHashRecordsReply'] = _FINDMISSINGHASHRECORDSREPLY
DESCRIPTOR.message_types_by_name['FindMissingSchemasRequest'] = _FINDMISSINGSCHEMASREQUEST
DESCRIPTOR.message_types_by_name['FindMissingSchemasReply'] = _FINDMISSINGSCHEMASREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ErrorProto = _reflection.GeneratedProtocolMessageType('ErrorProto', (_message.Message,), {
  'DESCRIPTOR' : _ERRORPROTO,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.ErrorProto)
  })
_sym_db.RegisterMessage(ErrorProto)

BranchRecord = _reflection.GeneratedProtocolMessageType('BranchRecord', (_message.Message,), {
  'DESCRIPTOR' : _BRANCHRECORD,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.BranchRecord)
  })
_sym_db.RegisterMessage(BranchRecord)

HashRecord = _reflection.GeneratedProtocolMessageType('HashRecord', (_message.Message,), {
  'DESCRIPTOR' : _HASHRECORD,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.HashRecord)
  })
_sym_db.RegisterMessage(HashRecord)

CommitRecord = _reflection.GeneratedProtocolMessageType('CommitRecord', (_message.Message,), {
  'DESCRIPTOR' : _COMMITRECORD,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.CommitRecord)
  })
_sym_db.RegisterMessage(CommitRecord)

SchemaRecord = _reflection.GeneratedProtocolMessageType('SchemaRecord', (_message.Message,), {
  'DESCRIPTOR' : _SCHEMARECORD,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.SchemaRecord)
  })
_sym_db.RegisterMessage(SchemaRecord)

PingRequest = _reflection.GeneratedProtocolMessageType('PingRequest', (_message.Message,), {
  'DESCRIPTOR' : _PINGREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PingRequest)
  })
_sym_db.RegisterMessage(PingRequest)

PingReply = _reflection.GeneratedProtocolMessageType('PingReply', (_message.Message,), {
  'DESCRIPTOR' : _PINGREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PingReply)
  })
_sym_db.RegisterMessage(PingReply)

GetClientConfigRequest = _reflection.GeneratedProtocolMessageType('GetClientConfigRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETCLIENTCONFIGREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.GetClientConfigRequest)
  })
_sym_db.RegisterMessage(GetClientConfigRequest)

GetClientConfigReply = _reflection.GeneratedProtocolMessageType('GetClientConfigReply', (_message.Message,), {

  'ConfigEntry' : _reflection.GeneratedProtocolMessageType('ConfigEntry', (_message.Message,), {
    'DESCRIPTOR' : _GETCLIENTCONFIGREPLY_CONFIGENTRY,
    '__module__' : 'hangar_service_pb2'
    # @@protoc_insertion_point(class_scope:hangar.GetClientConfigReply.ConfigEntry)
    })
  ,
  'DESCRIPTOR' : _GETCLIENTCONFIGREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.GetClientConfigReply)
  })
_sym_db.RegisterMessage(GetClientConfigReply)
_sym_db.RegisterMessage(GetClientConfigReply.ConfigEntry)

FetchBranchRecordRequest = _reflection.GeneratedProtocolMessageType('FetchBranchRecordRequest', (_message.Message,), {
  'DESCRIPTOR' : _FETCHBRANCHRECORDREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchBranchRecordRequest)
  })
_sym_db.RegisterMessage(FetchBranchRecordRequest)

FetchBranchRecordReply = _reflection.GeneratedProtocolMessageType('FetchBranchRecordReply', (_message.Message,), {
  'DESCRIPTOR' : _FETCHBRANCHRECORDREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchBranchRecordReply)
  })
_sym_db.RegisterMessage(FetchBranchRecordReply)

FetchDataRequest = _reflection.GeneratedProtocolMessageType('FetchDataRequest', (_message.Message,), {
  'DESCRIPTOR' : _FETCHDATAREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchDataRequest)
  })
_sym_db.RegisterMessage(FetchDataRequest)

FetchDataReply = _reflection.GeneratedProtocolMessageType('FetchDataReply', (_message.Message,), {
  'DESCRIPTOR' : _FETCHDATAREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchDataReply)
  })
_sym_db.RegisterMessage(FetchDataReply)

FetchCommitRequest = _reflection.GeneratedProtocolMessageType('FetchCommitRequest', (_message.Message,), {
  'DESCRIPTOR' : _FETCHCOMMITREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchCommitRequest)
  })
_sym_db.RegisterMessage(FetchCommitRequest)

FetchCommitReply = _reflection.GeneratedProtocolMessageType('FetchCommitReply', (_message.Message,), {
  'DESCRIPTOR' : _FETCHCOMMITREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchCommitReply)
  })
_sym_db.RegisterMessage(FetchCommitReply)

FetchSchemaRequest = _reflection.GeneratedProtocolMessageType('FetchSchemaRequest', (_message.Message,), {
  'DESCRIPTOR' : _FETCHSCHEMAREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchSchemaRequest)
  })
_sym_db.RegisterMessage(FetchSchemaRequest)

FetchSchemaReply = _reflection.GeneratedProtocolMessageType('FetchSchemaReply', (_message.Message,), {
  'DESCRIPTOR' : _FETCHSCHEMAREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchSchemaReply)
  })
_sym_db.RegisterMessage(FetchSchemaReply)

PushBranchRecordRequest = _reflection.GeneratedProtocolMessageType('PushBranchRecordRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUSHBRANCHRECORDREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushBranchRecordRequest)
  })
_sym_db.RegisterMessage(PushBranchRecordRequest)

PushBranchRecordReply = _reflection.GeneratedProtocolMessageType('PushBranchRecordReply', (_message.Message,), {
  'DESCRIPTOR' : _PUSHBRANCHRECORDREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushBranchRecordReply)
  })
_sym_db.RegisterMessage(PushBranchRecordReply)

PushDataRequest = _reflection.GeneratedProtocolMessageType('PushDataRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUSHDATAREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushDataRequest)
  })
_sym_db.RegisterMessage(PushDataRequest)

PushDataReply = _reflection.GeneratedProtocolMessageType('PushDataReply', (_message.Message,), {
  'DESCRIPTOR' : _PUSHDATAREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushDataReply)
  })
_sym_db.RegisterMessage(PushDataReply)

PushCommitRequest = _reflection.GeneratedProtocolMessageType('PushCommitRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUSHCOMMITREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushCommitRequest)
  })
_sym_db.RegisterMessage(PushCommitRequest)

PushCommitReply = _reflection.GeneratedProtocolMessageType('PushCommitReply', (_message.Message,), {
  'DESCRIPTOR' : _PUSHCOMMITREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushCommitReply)
  })
_sym_db.RegisterMessage(PushCommitReply)

PushSchemaRequest = _reflection.GeneratedProtocolMessageType('PushSchemaRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUSHSCHEMAREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushSchemaRequest)
  })
_sym_db.RegisterMessage(PushSchemaRequest)

PushSchemaReply = _reflection.GeneratedProtocolMessageType('PushSchemaReply', (_message.Message,), {
  'DESCRIPTOR' : _PUSHSCHEMAREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushSchemaReply)
  })
_sym_db.RegisterMessage(PushSchemaReply)

FindMissingCommitsRequest = _reflection.GeneratedProtocolMessageType('FindMissingCommitsRequest', (_message.Message,), {
  'DESCRIPTOR' : _FINDMISSINGCOMMITSREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FindMissingCommitsRequest)
  })
_sym_db.RegisterMessage(FindMissingCommitsRequest)

FindMissingCommitsReply = _reflection.GeneratedProtocolMessageType('FindMissingCommitsReply', (_message.Message,), {
  'DESCRIPTOR' : _FINDMISSINGCOMMITSREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FindMissingCommitsReply)
  })
_sym_db.RegisterMessage(FindMissingCommitsReply)

FindMissingHashRecordsRequest = _reflection.GeneratedProtocolMessageType('FindMissingHashRecordsRequest', (_message.Message,), {
  'DESCRIPTOR' : _FINDMISSINGHASHRECORDSREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FindMissingHashRecordsRequest)
  })
_sym_db.RegisterMessage(FindMissingHashRecordsRequest)

FindMissingHashRecordsReply = _reflection.GeneratedProtocolMessageType('FindMissingHashRecordsReply', (_message.Message,), {
  'DESCRIPTOR' : _FINDMISSINGHASHRECORDSREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FindMissingHashRecordsReply)
  })
_sym_db.RegisterMessage(FindMissingHashRecordsReply)

FindMissingSchemasRequest = _reflection.GeneratedProtocolMessageType('FindMissingSchemasRequest', (_message.Message,), {
  'DESCRIPTOR' : _FINDMISSINGSCHEMASREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FindMissingSchemasRequest)
  })
_sym_db.RegisterMessage(FindMissingSchemasRequest)

FindMissingSchemasReply = _reflection.GeneratedProtocolMessageType('FindMissingSchemasReply', (_message.Message,), {
  'DESCRIPTOR' : _FINDMISSINGSCHEMASREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FindMissingSchemasReply)
  })
_sym_db.RegisterMessage(FindMissingSchemasReply)


DESCRIPTOR._options = None
_GETCLIENTCONFIGREPLY_CONFIGENTRY._options = None

_HANGARSERVICE = _descriptor.ServiceDescriptor(
  name='HangarService',
  full_name='hangar.HangarService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=2356,
  serialized_end=3705,
  methods=[
  _descriptor.MethodDescriptor(
    name='PING',
    full_name='hangar.HangarService.PING',
    index=0,
    containing_service=None,
    input_type=_PINGREQUEST,
    output_type=_PINGREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetClientConfig',
    full_name='hangar.HangarService.GetClientConfig',
    index=1,
    containing_service=None,
    input_type=_GETCLIENTCONFIGREQUEST,
    output_type=_GETCLIENTCONFIGREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchBranchRecord',
    full_name='hangar.HangarService.FetchBranchRecord',
    index=2,
    containing_service=None,
    input_type=_FETCHBRANCHRECORDREQUEST,
    output_type=_FETCHBRANCHRECORDREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchData',
    full_name='hangar.HangarService.FetchData',
    index=3,
    containing_service=None,
    input_type=_FETCHDATAREQUEST,
    output_type=_FETCHDATAREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchCommit',
    full_name='hangar.HangarService.FetchCommit',
    index=4,
    containing_service=None,
    input_type=_FETCHCOMMITREQUEST,
    output_type=_FETCHCOMMITREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchSchema',
    full_name='hangar.HangarService.FetchSchema',
    index=5,
    containing_service=None,
    input_type=_FETCHSCHEMAREQUEST,
    output_type=_FETCHSCHEMAREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushBranchRecord',
    full_name='hangar.HangarService.PushBranchRecord',
    index=6,
    containing_service=None,
    input_type=_PUSHBRANCHRECORDREQUEST,
    output_type=_PUSHBRANCHRECORDREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushData',
    full_name='hangar.HangarService.PushData',
    index=7,
    containing_service=None,
    input_type=_PUSHDATAREQUEST,
    output_type=_PUSHDATAREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushCommit',
    full_name='hangar.HangarService.PushCommit',
    index=8,
    containing_service=None,
    input_type=_PUSHCOMMITREQUEST,
    output_type=_PUSHCOMMITREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushSchema',
    full_name='hangar.HangarService.PushSchema',
    index=9,
    containing_service=None,
    input_type=_PUSHSCHEMAREQUEST,
    output_type=_PUSHSCHEMAREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchFindMissingCommits',
    full_name='hangar.HangarService.FetchFindMissingCommits',
    index=10,
    containing_service=None,
    input_type=_FINDMISSINGCOMMITSREQUEST,
    output_type=_FINDMISSINGCOMMITSREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchFindMissingHashRecords',
    full_name='hangar.HangarService.FetchFindMissingHashRecords',
    index=11,
    containing_service=None,
    input_type=_FINDMISSINGHASHRECORDSREQUEST,
    output_type=_FINDMISSINGHASHRECORDSREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchFindMissingSchemas',
    full_name='hangar.HangarService.FetchFindMissingSchemas',
    index=12,
    containing_service=None,
    input_type=_FINDMISSINGSCHEMASREQUEST,
    output_type=_FINDMISSINGSCHEMASREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushFindMissingCommits',
    full_name='hangar.HangarService.PushFindMissingCommits',
    index=13,
    containing_service=None,
    input_type=_FINDMISSINGCOMMITSREQUEST,
    output_type=_FINDMISSINGCOMMITSREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushFindMissingHashRecords',
    full_name='hangar.HangarService.PushFindMissingHashRecords',
    index=14,
    containing_service=None,
    input_type=_FINDMISSINGHASHRECORDSREQUEST,
    output_type=_FINDMISSINGHASHRECORDSREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushFindMissingSchemas',
    full_name='hangar.HangarService.PushFindMissingSchemas',
    index=15,
    containing_service=None,
    input_type=_FINDMISSINGSCHEMASREQUEST,
    output_type=_FINDMISSINGSCHEMASREPLY,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_HANGARSERVICE)

DESCRIPTOR.services_by_name['HangarService'] = _HANGARSERVICE

# @@protoc_insertion_point(module_scope)



"""Interceptor that ensures a specific header is present.

Portions of this code have been taken and modified from the "gRPC" project.

URL:      https://github.com/grpc/grpc/
File:     examples/python/interceptors/default_value/default_value_client_interceptor.py
Commit:   6146151a4fe1e28921c12d1ae5635e113a24b9d7
Accessed: 23 APR 2019

gRPC License
-------------------------------------------------------------------------------
Copyright 2017 gRPC authors.

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
from os.path import split

import grpc


SERVICE_METHOD_TYPES = {
    'PING': 'uu',
    'GetClientConfig': 'uu',
    'FetchBranchRecord': 'uu',
    'FetchData': 'ss',
    'FetchCommit': 'us',
    'FetchSchema': 'uu',
    'PushBranchRecord': 'uu',
    'PushData': 'su',
    'PushCommit': 'su',
    'PushSchema': 'uu',
    'FetchFindMissingCommits': 'uu',
    'FetchFindMissingHashRecords': 'ss',
    'FetchFindMissingSchemas': 'uu',
    'PushFindMissingCommits': 'uu',
    'PushFindMissingHashRecords': 'ss',
    'PushFindMissingSchemas': 'uu',
}


def _unary_unary_rpc_terminator(code, details):
    def terminate(ignored_request, context):
        context.abort(code, details)
    return grpc.unary_unary_rpc_method_handler(terminate)


def _unary_stream_rpc_terminator(code, details):  # pragma: no cover
    def terminate(ignored_request, context):
        context.abort(code, details)
    return grpc.unary_stream_rpc_method_handler(terminate)


def _stream_unary_rpc_terminator(code, details):  # pragma: no cover
    def terminate(ignored_request, context):
        context.abort(code, details)
    return grpc.unary_stream_rpc_method_handler(terminate)


def _stream_stream_rpc_terminator(code, details):  # pragma: no cover
    def terminate(ignored_request, context):
        context.abort(code, details)
    return grpc.stream_stream_rpc_method_handler(terminate)


def _select_rpc_terminator(intercepted_method):
    method_type = SERVICE_METHOD_TYPES[intercepted_method]

    if method_type == 'uu':
        return _unary_unary_rpc_terminator
    elif method_type == 'su':  # pragma: no cover
        return _stream_unary_rpc_terminator
    elif method_type == 'us':  # pragma: no cover
        return _unary_stream_rpc_terminator
    elif method_type == 'ss':  # pragma: no cover
        return _stream_stream_rpc_terminator
    else:                      # pragma: no cover
        raise ValueError(f'unknown method type: {method_type} for service: {intercepted_method}')


class RequestHeaderValidatorInterceptor(grpc.ServerInterceptor):

    def __init__(self, push_restricted, header, value, code, details):
        self._push_restricted = push_restricted
        self._header = header
        self._value = value
        self._code = code
        self._details = details

    def intercept_service(self, continuation, handler_call_details):
        _, intercepted_method = split(handler_call_details.method)
        print(f'intercepted method: {intercepted_method}')

        if (intercepted_method.startswith('Push') is True) and (self._push_restricted is True):
            if (self._header, self._value) in handler_call_details.invocation_metadata:
                return continuation(handler_call_details)
            else:
                return _select_rpc_terminator(intercepted_method)(self._code, self._details)
        else:
            return continuation(handler_call_details)



import warnings
from collections import namedtuple
from typing import Sequence

from .common import GroupedColumns
from ..utils import LazyLoader


try:
    torchdata = LazyLoader('torchdata', globals(), 'torch.utils.data')
except (ImportError, ModuleNotFoundError):
    raise ImportError(
        'Could not import "pytorch" library. Ensure library is '
        'installed correctly to use pytorch dataloader functions')


def make_torch_dataset(columns,
                       keys: Sequence[str] = None,
                       index_range: slice = None,
                       field_names: Sequence[str] = None):
    """
    Returns a :class:`torch.utils.data.Dataset` object which can be loaded into
    a :class:`torch.utils.data.DataLoader`.

    .. warning::

       On Windows systems, setting the parameter ``num_workers`` in the
       resulting :class:`torch.utils.data.DataLoader` method will result in a
       RuntimeError or deadlock. This is due to limitations of multiprocess
       start methods on Windows itself. Using the default argument value
       (``num_workers=0``) will let the DataLoader work in single process mode
       as expected.

    Parameters
    ----------
    columns : :class:`~hangar.columns.column.Columns` or Sequence
        A column object, a tuple of column object or a list of column
        objects.
    keys : Sequence[str]
        An iterable collection of sample names. If given only those samples will
        fetched from the column
    index_range : slice
        A python slice object which will be used to find the subset of column.
        Argument `keys` takes priority over `range` i.e. if both are given, keys
        will be used and `range` will be ignored
    field_names : Sequence[str], optional
        An array of field names used as the `field_names` for the returned
        dict keys. If not given, column names will be used as the field_names.

    Examples
    --------
    >>> from hangar import Repository
    >>> from torch.utils.data import DataLoader
    >>> from hangar import make_torch_dataset
    >>> repo = Repository('.')
    >>> co = repo.checkout()
    >>> aset = co.columns['dummy_aset']
    >>> torch_dset = make_torch_dataset(aset, index_range=slice(1, 100))
    >>> loader = DataLoader(torch_dset, batch_size=16)
    >>> for batch in loader:
    ...     train_model(batch)

    Returns
    -------
    :class:`torch.utils.data.Dataset`
    """
    warnings.warn("Dataloaders are experimental in the current release.", UserWarning)
    if keys:
        if not isinstance(keys, (list, tuple, set)):
            raise TypeError(f'type(keys): {type(keys)} != (list, tuple, set)')

    gcols = GroupedColumns(columns, keys, index_range)
    if field_names:
        if not isinstance(field_names, (list, tuple, set)):
            raise TypeError(f'type(field_names): {type(field_names)} not collection')
        if len(field_names) != len(columns):
            err = f'# field_names {len(field_names)} != # columns: {len(columns)}'
            raise ValueError(err)
        BTName = '_'.join(['BatchTuple', *field_names])
        BTFieldNames = field_names
    else:
        BTName = '_'.join(['BatchTuple', *gcols.column_names])
        BTFieldNames = gcols.column_names

    wrapper = namedtuple(BTName, field_names=BTFieldNames, rename=True)
    globals()[BTName] = wrapper
    return TorchDataset(hangar_columns=gcols.columns_col,
                        sample_names=gcols.sample_names,
                        wrapper=wrapper)


class TorchDataset(torchdata.Dataset):
    """A wrapper around torch Dataset

    TorchDataset inherits :class:`torch.utils.data.Dataset` and accepts few
    convenient arguments to wrap hangar columns to be used in
    :class:`torch.utils.data.DataLoaders`.

    .. note::

        From PyTorch 1.1 onwards, if Dataset returns dict, DataLoader also
        returns dict

    Parameters
    ----------
    columns : :class:`~hangar.columns.column.Columns` or Sequence
        A list/tuple of hangar_arrayset objects with same length and contains
        same keys. This class doesn't do any explicit check for length or the
        key names and assumes those all the columns are valid as per the
        requirement
    sample_names : tuple of allowed sample names/keys
        User can select a subset of all the available samples and pass the
        names for only those
    wrapper : namedtuple
        namedtuple placed in global memory used to wrap the output from
        __getitem__
    """

    def __init__(self, hangar_columns, sample_names, wrapper):
        self.hangar_columns = hangar_columns
        self.sample_names = sample_names
        self.wrapper: namedtuple = wrapper

    def __len__(self) -> int:
        """
        Length of the available and allowed samples

        Returns
        -------
        int
            number of samples retrievable by the dataloader.
        """
        return len(self.sample_names)

    def __getitem__(self, index: int):
        """Use data names array to find the sample name at an index and loop
        through the array of hangar columns to return the sample.

        Parameters
        ----------
        index : int
            some sample index location.

        Returns
        -------
        namedtuple[:class:`torch.Tensor`]
            One sample with the given name from all the provided columns
        """
        key = self.sample_names[index]
        out = []
        for aset in self.hangar_columns:
            out.append(aset.get(key))
        return self.wrapper._make(out)



from typing import Dict, List, Tuple, Union


required_packages = [
    ('hangar', lambda p: p.__version__),
    ('click', lambda p: p.__version__),
    ('lmdb', lambda p: p.__version__),
    ('h5py', lambda p: p.__version__),
    ('hdf5plugin', lambda p: p.version),
    ('numpy', lambda p: p.__version__),
    ('blosc', lambda p: p.__version__),
    ('tqdm', lambda p: p.__version__),
    ('wrapt', lambda p: p.__version__),
    ('grpc', lambda p: p.__version__),
    ('xxhash', lambda p: p.VERSION),
]


def get_versions() -> dict:
    """Return information on software, machine, installed versions of packages.

    dict
        host, package, and `optional` package info.
    """
    d = {'host': get_system_info(),
         'packages': get_package_info(required_packages),
         'optional': get_optional_info()}
    return d


def get_system_info() -> List[Tuple[str, str]]:
    """Return local computer python, OS, and Machine info

    Returns
    -------
    List[Tuple[str, str]]
        field collected and value of the system parameter.
    """
    import locale
    import os
    import platform
    import struct
    import sys

    (sysname, nodename, release,
     version, machine, processor) = platform.uname()

    try:
        loc = locale.getlocale()
    except ValueError:  # pragma: no cover
        loc = None

    host = [
        ('python', f'{sys.version_info[:]}'),
        ('python-bits', f'{struct.calcsize("P") * 8}'),
        ('OS', f'{sysname}'),
        ('OS-release', f'{release}'),
        ('machine', f'{machine}'),
        ('processor', f'{processor}'),
        ('byteorder', f'{sys.byteorder}'),
        ('LC_ALL', f'{os.environ.get("LC_ALL", "None")}'),
        ('LANG', f'{os.environ.get("LANG", "None")}'),
        ('LOCALE', f'{loc}'),
        ('cpu-count', f'{os.cpu_count()}'),
    ]

    return host


def get_optional_info() -> Dict[str, Union[str, bool]]:
    """Get optional package info (tensorflow, pytorch, hdf5_bloscfilter, etc.)

    Returns
    -------
    Dict[str, Union[str, False]]
        package name, package version (if installed, otherwise False)
    """
    res = {}
    try:
        import h5py
        bloscFilterAvail = h5py.h5z.filter_avail(32001)
    except ImportError:  # pragma: no cover
        bloscFilterAvail = False
    res['blosc-hdf5-plugin'] = bloscFilterAvail

    try:
        import torch
        torchVersion = torch.__version__
    except ImportError:  # pragma: no cover
        torchVersion = False
    res['pytorch'] = torchVersion

    try:
        import tensorflow
        tensorflowVersion = tensorflow.__version__
    except ImportError:  # pragma: no cover
        tensorflowVersion = False
    res['tensorflow'] = tensorflowVersion

    return res


def get_package_info(pkgs):
    """ get package versions for the passed required & optional packages.

    Using local imports to avoid import overhead on interpreter startup.
    """
    import importlib

    pversions = []
    for modname, ver_f in pkgs:
        try:
            mod = importlib.import_module(modname)
            ver = ver_f(mod)
            pversions.append((modname, ver))
        except Exception:  # pragma: no cover
            pversions.append((modname, None))

    return pversions



"""Interceptor that adds headers to outgoing requests

Portions of this code have been taken and modified from the "gRPC" project.

URL:      https://github.com/grpc/grpc/
File:     examples/python/interceptors/default_value/header_manipulator_client_interceptor.py
Commit:   87cd994b0477e98c976e7b321b3c1f52666ab5e0
Accessed: 23 APR 2019

gRPC License
-------------------------------------------------------------------------------
Copyright 2017 gRPC authors.

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
from collections import namedtuple

import grpc


class _GenericClientInterceptor(
        grpc.UnaryUnaryClientInterceptor, grpc.UnaryStreamClientInterceptor,
        grpc.StreamUnaryClientInterceptor, grpc.StreamStreamClientInterceptor):
    """Base class for interceptors that operate on all RPC types."""

    def __init__(self, interceptor_function):
        self._fn = interceptor_function

    def intercept_unary_unary(self, continuation, client_call_details, request):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, False)
        response = continuation(new_details, next(new_request_iterator))
        return postprocess(response) if postprocess else response

    def intercept_unary_stream(self, continuation, client_call_details,
                               request):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, True)
        response_it = continuation(new_details, next(new_request_iterator))
        return postprocess(response_it) if postprocess else response_it

    def intercept_stream_unary(self, continuation, client_call_details,
                               request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator, True, False)
        response = continuation(new_details, new_request_iterator)
        return postprocess(response) if postprocess else response

    def intercept_stream_stream(self, continuation, client_call_details,
                                request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator, True, True)
        response_it = continuation(new_details, new_request_iterator)
        return postprocess(response_it) if postprocess else response_it


def create_client_interceptor(intercept_call):
    return _GenericClientInterceptor(intercept_call)


class _ClientCallDetails(
        namedtuple(
            typename='_ClientCallDetails',
            field_names=('method', 'timeout', 'metadata', 'credentials')),
        grpc.ClientCallDetails):
    pass


def header_adder_interceptor(header, value):
    """Interceptor that adds headers to outgoing requests."""

    def intercept_call(client_call_details, request_iterator, request_streaming,
                       response_streaming):
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)

        if (header != '') and (value != ''):
            metadata.append((header, value))
        client_call_details = _ClientCallDetails(
            client_call_details.method,
            client_call_details.timeout,
            metadata,
            client_call_details.credentials)
        return client_call_details, request_iterator, None

    return create_client_interceptor(intercept_call)




class CheckoutDictIteration:
    """Mixin class for checkout objects which mock common iter methods

    Methods
    -------
    __len__
    __contains__
    __iter__
    keys
    values
    items
    """

    def __len__(self):
        """Returns number of columns in the checkout.
        """
        self._verify_alive()
        return len(self.columns)

    def __contains__(self, key):
        """Determine if some column name (key) exists in the checkout.
        """
        self._verify_alive()
        return bool(key in self.columns)

    def __iter__(self):
        """Iterate over column keys"""
        self._verify_alive()
        return iter(self.columns)

    def keys(self):
        """Generator yielding the name (key) of every column
        """
        self._verify_alive()
        yield from self.columns.keys()

    def values(self):
        """Generator yielding accessor object of every column
        """
        self._verify_alive()
        yield from self.columns.values()

    def items(self):
        """Generator yielding tuple of (name, accessor object) of every column
        """
        self._verify_alive()
        yield from self.columns.items()



print("import: 'hangar'")
import hangar

print("import: 'hangar.diagnostics'")
import hangar.diagnostics

print("import: 'hangar.records'")
import hangar.records

print("import: 'hangar.remote'")
import hangar.remote

print("import: 'hangar.backends'")
import hangar.backends

print("import: 'hangar.columns'")
import hangar.columns




# -*- coding: utf-8 -*-
"""
Portions of this code have been taken and modified from the "asciidag" project.

URL:      https://github.com/sambrightman/asciidag/
File:     asciidag/graph.py
Commit:   7c1eefe3895630dc3906bbe9d553e0169202756a
Accessed: 25 MAR 2019

asciidag License
-------------------------------------------------------------------------------
License: Mozilla Public License 2.0
URL:     https://github.com/sambrightman/asciidag/blob/7c1eefe3895630dc3906bbe9d553e0169202756a/LICENSE
"""

import sys
import time
from enum import Enum

__all__ = ('Graph',)

COLOR_NORMAL = ""
COLOR_RESET = "\033[m"
COLOR_BOLD = "\033[1m"
COLOR_RED = "\033[31m"
COLOR_GREEN = "\033[32m"
COLOR_YELLOW = "\033[33m"
COLOR_BLUE = "\033[34m"
COLOR_MAGENTA = "\033[35m"
COLOR_CYAN = "\033[36m"
COLOR_BOLD_RED = "\033[1;31m"
COLOR_BOLD_GREEN = "\033[1;32m"
COLOR_BOLD_YELLOW = "\033[1;33m"
COLOR_BOLD_BLUE = "\033[1;34m"
COLOR_BOLD_MAGENTA = "\033[1;35m"
COLOR_BOLD_CYAN = "\033[1;36m"
COLOR_BG_RED = "\033[41m"
COLOR_BG_GREEN = "\033[42m"
COLOR_BG_YELLOW = "\033[43m"
COLOR_BG_BLUE = "\033[44m"
COLOR_BG_MAGENTA = "\033[45m"
COLOR_BG_CYAN = "\033[46m"

COLUMN_COLORS_ANSI = [
    COLOR_BOLD_RED,
    COLOR_BOLD_GREEN,
    COLOR_BOLD_YELLOW,
    COLOR_BOLD_BLUE,
    COLOR_BOLD_MAGENTA,
    COLOR_BOLD_CYAN,
    COLOR_RED,
    COLOR_GREEN,
    COLOR_YELLOW,
    COLOR_BLUE,
    COLOR_MAGENTA,
    COLOR_CYAN,
    COLOR_RESET,
]


class Column(object):  # pylint: disable=too-few-public-methods
    """A single column of output.

    Attributes:
        commit -- The parent commit of this column.
        color  -- The color to (optionally) print this column in.
                  This is an index into column_colors.

    """

    def __init__(self, commit, color):
        self.commit = commit
        self.color = color


class GraphState(Enum):  # pylint: disable=too-few-public-methods
    PADDING = 0
    SKIP = 1
    PRE_COMMIT = 2
    COMMIT = 3
    POST_MERGE = 4
    COLLAPSING = 5


class Graph(object):  # pragma: no cover
    """
    The commit currently being processed
        struct commit *commit

    The number of interesting parents that this commit has. Note that this is
    not the same as the actual number of parents. This count excludes parents
    that won't be printed in the graph output, as determined by
    is_interesting().
        int num_parents

    The width of the graph output for this commit. All rows for this commit are
    padded to this width, so that messages printed after the graph output are
    aligned.
        int width

    The next expansion row to print when state is GraphState.PRE_COMMIT
        int expansion_row

    The current output state. This tells us what kind of line next_line()
    should output.
        enum graph_state state

    The output state for the previous line of output. This is primarily used to
    determine how the first merge line should appear, based on the last line of
    the previous commit.
        enum graph_state prev_state

    The index of the column that refers to this commit. If none of the incoming
    columns refer to this commit, this will be equal to num_columns.
        int commit_index

    The commit_index for the previously displayed commit. This is used to
    determine how the first line of a merge graph output should appear, based
    on the last line of the previous commit.
        int prev_commit_index

    The maximum number of columns that can be stored in the columns and
    new_columns arrays. This is also half the number of entries that can be
    stored in the mapping and new_mapping arrays.
        int column_capacity

    The number of columns (also called "branch lines" in some places)
        int num_columns

    The number of columns in the new_columns array
        int num_new_columns

    The number of entries in the mapping array
        int mapping_size

    The column state before we output the current commit.
        struct column *columns

    The new column state after we output the current commit. Only valid when
    state is GraphState.COLLAPSING.
        struct column *new_columns

    An array that tracks the current state of each character in the output line
    during state GraphState.COLLAPSING. Each entry is -1 if this character is
    empty, or a non-negative integer if the character contains a branch line.
    The value of the integer indicates the target position for this branch
    line. (I.e., this array maps the current column positions to their desired
    positions.)

    The maximum capacity of this array is always sizeof(int) * 2 *
    column_capacity.
        int *mapping

    A temporary array for computing the next mapping state while we are
    outputting a mapping line. This is stored as part of the git_graph simply
    so we don't have to allocate a new temporary array each time we have to
    output a collapsing line.
        int *new_mapping

    The current default column color being used. This is stored as an index
    into the array column_colors.
        unsigned short default_column_color
    """
    def __init__(self,
                 fh=None,
                 first_parent_only=False,
                 use_color=True,
                 column_colors=None):
        """State machine for processing DAG nodes into ASCII graphs.

        show_nodes() deals with sorting the nodes from tips down into
        topological order. It then displays them line-by-line.

        """
        self.commit = None
        self.buf = ''

        if fh is None:
            self.outfile = sys.stdout
        else:
            self.outfile = fh
        self.first_parent_only = first_parent_only
        self.use_color = use_color
        if column_colors is None:
            self.column_colors = COLUMN_COLORS_ANSI
        else:
            self.column_colors = column_colors

        self.num_parents = 0
        self.width = 0
        self.expansion_row = 0
        self.state = GraphState.PADDING
        self.prev_state = GraphState.PADDING
        self.commit_index = 0
        self.prev_commit_index = 0
        self.num_columns = 0
        self.num_new_columns = 0
        self.mapping_size = 0
        # Start the column color at the maximum value, since we'll always
        # increment it for the first commit we output. This way we start at 0
        # for the first commit.
        self.default_column_color = len(self.column_colors) - 1

        self.columns = {}
        self.new_columns = {}
        self.mapping = {}
        self.new_mapping = {}

    def show_nodes(self, dag, spec, branch, start, order, stop='',
                   *, show_time=True, show_user=True):
        """Printing function that displays a DAG representing the commit history

        Print a revision history alongside a revision graph drawn with ASCII
        characters. Nodes printed as an * character are parents of the working
        directory. Any unreachable (but referenced nodes) are displayed at +

        Parameters
        ----------
        dag : dict
            directed acyclic graph of nodes and connections in commits. No more than
            2 connections per node
        spec: dict
            dictionary of commit specification (user name, email, message, etc).
        branch : dict
            dict of commit hash -> list of branch names whose HEAD commit is at
            that key.
        start : string
            commit hash to act as the top of the topological sort.
        order: list
            time based ordering of commit hashs
        stop : str, optional
            commit hash to stop generating the graph at if the DAG contains more
            history than is needed (the default is '', which is the "parent" of
            the initial repository commit.)
        """
        if start == stop:
            return

        fmtSpec = {}
        for cmt, cmtspec in spec.items():
            if show_time:
                t = f"({time.strftime('%d%b%Y %H:%M:%S', time.gmtime(cmtspec['commit_time']))})"
            else:
                t = ''
            if show_user:
                u = f"({cmtspec['commit_user']})"
            else:
                u = ''
            m = cmtspec['commit_message']
            br = ' '
            if cmt in branch:
                for branchName in branch[cmt]:
                    if self.use_color is True:
                        br = f'{br}({COLOR_BOLD_RED}{branchName}{COLOR_RESET}) '
                    else:
                        br = f'{br}({branchName}) '
            fmtSpec[cmt] = f'{cmt}{br}{t}{u}: {m}'

        for rev in order:
            parents = dag[rev]
            self._update(rev, parents)
            self._show_commit()
            self.outfile.write(fmtSpec[rev])
            if not self._is_commit_finished():
                self.outfile.write('\n')
                self._show_remainder()
            self.outfile.write('\n')

    def _write_column(self, col, col_char):
        if col.color is not None:
            self.buf += self.column_colors[col.color]
        self.buf += col_char
        if col.color is not None:
            self.buf += self.column_colors[-1]

    def _update_state(self, state):
        self.prev_state = self.state
        self.state = state

    def _interesting_parents(self):
        for parent in self.commit_parents:
            yield parent
            if self.first_parent_only:
                break

    def _get_current_column_color(self):
        if not self.use_color:
            return None
        return self.default_column_color

    def _increment_column_color(self):
        self.default_column_color = ((self.default_column_color + 1)
                                     % len(self.column_colors))

    def _find_commit_color(self, commit):
        for i in range(self.num_columns):
            if self.columns[i].commit == commit:
                return self.columns[i].color
        return self._get_current_column_color()

    def _insert_into_new_columns(self, commit, mapping_index):
        """
        If the commit is already in the new_columns list, we don't need to add
        it. Just update the mapping correctly.
        """
        for i in range(self.num_new_columns):
            if self.new_columns[i].commit == commit:
                self.mapping[mapping_index] = i
                return mapping_index + 2

        # This commit isn't already in new_columns. Add it.
        column = Column(commit, self._find_commit_color(commit))
        self.new_columns[self.num_new_columns] = column
        self.mapping[mapping_index] = self.num_new_columns
        self.num_new_columns += 1
        return mapping_index + 2

    def _update_width(self, is_commit_in_existing_columns):
        """
        Compute the width needed to display the graph for this commit. This is
        the maximum width needed for any row. All other rows will be padded to
        this width.

        Compute the number of columns in the widest row: Count each existing
        column (self.num_columns), and each new column added by this commit.
        """
        max_cols = self.num_columns + self.num_parents

        # Even if the current commit has no parents to be printed, it still
        # takes up a column for itself.
        if self.num_parents < 1:
            max_cols += 1

        # We added a column for the current commit as part of self.num_parents.
        # If the current commit was already in self.columns, then we have double
        # counted it.
        if is_commit_in_existing_columns:
            max_cols -= 1

        # Each column takes up 2 spaces
        self.width = max_cols * 2

    def _update_columns(self):
        """
        Swap self.columns with self.new_columns self.columns contains the state
        for the previous commit, and new_columns now contains the state for our
        commit.

        We'll re-use the old columns array as storage to compute the new columns
        list for the commit after this one.
        """
        self.columns, self.new_columns = self.new_columns, self.columns
        self.num_columns = self.num_new_columns
        self.num_new_columns = 0

        # Now update new_columns and mapping with the information for the commit
        # after this one.
        #
        # First, make sure we have enough room. At most, there will be
        # self.num_columns + self.num_parents columns for the next commit.
        max_new_columns = self.num_columns + self.num_parents

        # Clear out self.mapping
        self.mapping_size = 2 * max_new_columns
        for i in range(self.mapping_size):
            self.mapping[i] = -1

        # Populate self.new_columns and self.mapping
        #
        # Some of the parents of this commit may already be in self.columns. If
        # so, self.new_columns should only contain a single entry for each such
        # commit. self.mapping should contain information about where each
        # current branch line is supposed to end up after the collapsing is
        # performed.
        seen_this = False
        mapping_idx = 0
        is_commit_in_columns = True
        for i in range(self.num_columns + 1):
            if i == self.num_columns:
                if seen_this:
                    break
                is_commit_in_columns = False
                col_commit = self.commit
            else:
                col_commit = self.columns[i].commit

            if col_commit == self.commit:
                old_mapping_idx = mapping_idx
                seen_this = True
                self.commit_index = i
                for parent in self._interesting_parents():
                    # If this is a merge, or the start of a new childless
                    # column, increment the current color.
                    if self.num_parents > 1 or not is_commit_in_columns:
                        self._increment_column_color()
                    mapping_idx = self._insert_into_new_columns(
                        parent,
                        mapping_idx)
                # We always need to increment mapping_idx by at least 2, even if
                # it has no interesting parents. The current commit always takes
                # up at least 2 spaces.
                if mapping_idx == old_mapping_idx:
                    mapping_idx += 2
            else:
                mapping_idx = self._insert_into_new_columns(col_commit,
                                                            mapping_idx)

        # Shrink mapping_size to be the minimum necessary
        while (self.mapping_size > 1 and
               self.mapping[self.mapping_size - 1] < 0):
            self.mapping_size -= 1

        # Compute self.width for this commit
        self._update_width(is_commit_in_columns)

    def _update(self, commit, parents):
        self.commit = commit
        self.commit_parents = parents
        self.num_parents = len(list(self._interesting_parents()))

        # Store the old commit_index in prev_commit_index.
        # update_columns() will update self.commit_index for this commit.
        self.prev_commit_index = self.commit_index

        # Call update_columns() to update
        # columns, new_columns, and mapping.
        self._update_columns()
        self.expansion_row = 0

        # Update self.state.
        # Note that we don't call update_state() here, since we don't want to
        # update self.prev_state. No line for self.state was ever printed.
        #
        # If the previous commit didn't get to the GraphState.PADDING state, it
        # never finished its output. Goto GraphState.SKIP, to print out a line
        # to indicate that portion of the graph is missing.
        #
        # If there are 3 or more parents, we may need to print extra rows before
        # the commit, to expand the branch lines around it and make room for it.
        # We need to do this only if there is a branch row (or more) to the
        # right of this commit.
        #
        # If less than 3 parents, we can immediately print the commit line.
        if self.state != GraphState.PADDING:
            self.state = GraphState.SKIP
        elif (self.num_parents >= 3 and
              self.commit_index < (self.num_columns - 1)):
            self.state = GraphState.PRE_COMMIT  # noqa: E501 pylint: disable=redefined-variable-type
        else:
            self.state = GraphState.COMMIT

    def _is_mapping_correct(self):
        """
        The mapping is up to date if each entry is at its target, or is 1
        greater than its target. (If it is 1 greater than the target, '/' will
        be printed, so it will look correct on the next row.)
        """
        for i in range(self.mapping_size):
            target = self.mapping[i]
            if target < 0:
                continue
            if target == i // 2:
                continue
            return False
        return True

    def _pad_horizontally(self, chars_written):
        """Add spaces to string end so all lines of a commit have the same width.

        This way, fields printed to the right of the graph will remain aligned
        for the entire commit.
        """
        if chars_written >= self.width:
            return

        extra = self.width - chars_written
        self.buf += ' ' * extra

    def _output_padding_line(self):
        """Output a padding row, that leaves all branch lines unchanged
        """
        for i in range(self.num_new_columns):
            self._write_column(self.new_columns[i], '|')
            self.buf += ' '

        self._pad_horizontally(self.num_new_columns * 2)

    def _output_skip_line(self):
        """Output an ellipsis to indicate that a portion of the graph is missing.
        """
        self.buf += '...'
        self._pad_horizontally(3)

        if self.num_parents >= 3 and self.commit_index < self.num_columns - 1:
            self._update_state(GraphState.PRE_COMMIT)
        else:
            self._update_state(GraphState.COMMIT)

    def _output_pre_commit_line(self):
        """Formats a row with increased space around a commit with multiple parents.

        This is done in order to make room for the commit. It should only be
        called when there are 3 or more parents. We need 2 extra rows for every
        parent over 2.
        """
        assert self.num_parents >= 3, 'not enough parents to add expansion row'
        num_expansion_rows = (self.num_parents - 2) * 2

        # self.expansion_row tracks the current expansion row we are on.
        # It should be in the range [0, num_expansion_rows - 1]
        assert (0 <= self.expansion_row < num_expansion_rows), \
            'wrong number of expansion rows'

        # Output the row
        seen_this = False
        chars_written = 0
        for i in range(self.num_columns):
            col = self.columns[i]
            if col.commit == self.commit:
                seen_this = True
                self._write_column(col, '|')
                self.buf += ' ' * self.expansion_row
                chars_written += 1 + self.expansion_row
            elif seen_this and (self.expansion_row == 0):
                # This is the first line of the pre-commit output. If the
                # previous commit was a merge commit and ended in the
                # GraphState.POST_MERGE state, all branch lines after
                # self.prev_commit_index were printed as "\" on the previous
                # line. Continue to print them as "\" on this line. Otherwise,
                # print the branch lines as "|".
                if (self.prev_state == GraphState.POST_MERGE and
                        self.prev_commit_index < i):
                    self._write_column(col, '\\')
                else:
                    self._write_column(col, '|')
                chars_written += 1
            elif seen_this and (self.expansion_row > 0):
                self._write_column(col, '\\')
                chars_written += 1
            else:
                self._write_column(col, '|')
                chars_written += 1
            self.buf += ' '
            chars_written += 1

        self._pad_horizontally(chars_written)

        # Increment self.expansion_row, and move to state GraphState.COMMIT if
        # necessary
        self.expansion_row += 1
        if self.expansion_row >= num_expansion_rows:
            self._update_state(GraphState.COMMIT)

    # Draw an octopus merge and return the number of characters written.
    def _draw_octopus_merge(self):
        """
        Here dashless_commits represents the number of parents which don't
        need to have dashes (because their edges fit neatly under the commit).
        """
        dashless_commits = 2
        num_dashes = ((self.num_parents - dashless_commits) * 2) - 1
        for i in range(num_dashes):
            col_num = i // 2 + dashless_commits + self.commit_index
            self._write_column(self.new_columns[col_num], '-')
        col_num = num_dashes // 2 + dashless_commits + self.commit_index
        self._write_column(self.new_columns[col_num], '.')
        return num_dashes + 1

    def _output_commit_line(self):  # noqa: C901, E501 pylint: disable=too-many-branches
        """
        Output the row containing this commit Iterate up to and including
        self.num_columns, since the current commit may not be in any of the
        existing columns. (This happens when the current commit doesn't have
        any children that we have already processed.)
        """
        seen_this = False
        chars_written = 0
        for i in range(self.num_columns + 1):
            if i == self.num_columns:
                if seen_this:
                    break
                col_commit = self.commit
            else:
                col = self.columns[i]
                col_commit = self.columns[i].commit

            if col_commit == self.commit:
                seen_this = True
                self.buf += '*'
                chars_written += 1

                if self.num_parents > 2:
                    chars_written += self._draw_octopus_merge()
            elif seen_this and self.num_parents > 2:
                self._write_column(col, '\\')
                chars_written += 1
            elif seen_this and self.num_parents == 2:
                # This is a 2-way merge commit. There is no
                # GraphState.PRE_COMMIT stage for 2-way merges, so this is the
                # first line of output for this commit. Check to see what the
                # previous line of output was.
                #
                # If it was GraphState.POST_MERGE, the branch line coming into
                # this commit may have been '\', and not '|' or '/'. If so,
                # output the branch line as '\' on this line, instead of '|'.
                # This makes the output look nicer.
                if (self.prev_state == GraphState.POST_MERGE and
                        self.prev_commit_index < i):
                    self._write_column(col, '\\')
                else:
                    self._write_column(col, '|')
                chars_written += 1
            else:
                self._write_column(col, '|')
                chars_written += 1
            self.buf += ' '
            chars_written += 1

        self._pad_horizontally(chars_written)
        if self.num_parents > 1:
            self._update_state(GraphState.POST_MERGE)
        elif self._is_mapping_correct():
            self._update_state(GraphState.PADDING)
        else:
            self._update_state(GraphState.COLLAPSING)

    def _find_new_column_by_commit(self, commit):
        for i in range(self.num_new_columns):
            if self.new_columns[i].commit == commit:
                return self.new_columns[i]
        return None

    def _output_post_merge_line(self):
        seen_this = False
        chars_written = 0
        for i in range(self.num_columns + 1):
            if i == self.num_columns:
                if seen_this:
                    break
                col_commit = self.commit
            else:
                col = self.columns[i]
                col_commit = col.commit

            if col_commit == self.commit:
                # Since the current commit is a merge find the columns for the
                # parent commits in new_columns and use those to format the
                # edges.
                seen_this = True
                parents = self._interesting_parents()
                assert parents, 'merge has no parents'
                par_column = self._find_new_column_by_commit(next(parents))
                assert par_column, 'parent column not found'
                self._write_column(par_column, '|')
                chars_written += 1
                for parent in parents:
                    assert parent, 'parent is not valid'
                    par_column = self._find_new_column_by_commit(parent)
                    assert par_column, 'parent column not found'
                    self._write_column(par_column, '\\')
                    self.buf += ' '
                chars_written += (self.num_parents - 1) * 2
            elif seen_this:
                self._write_column(col, '\\')
                self.buf += ' '
                chars_written += 2
            else:
                self._write_column(col, '|')
                self.buf += ' '
                chars_written += 2

        self._pad_horizontally(chars_written)

        if self._is_mapping_correct():
            self._update_state(GraphState.PADDING)
        else:
            self._update_state(GraphState.COLLAPSING)

    def _output_collapsing_line(self):  # noqa: C901, E501 pylint: disable=too-many-branches
        used_horizontal = False
        horizontal_edge = -1
        horizontal_edge_target = -1

        # Clear out the new_mapping array
        for i in range(self.mapping_size):
            self.new_mapping[i] = -1

        for i in range(self.mapping_size):
            target = self.mapping[i]
            if target < 0:
                continue

            # Since update_columns() always inserts the leftmost column first,
            # each branch's target location should always be either its current
            # location or to the left of its current location.
            #
            # We never have to move branches to the right. This makes the graph
            # much more legible, since whenever branches cross, only one is
            # moving directions.
            assert target * 2 <= i, \
                'position {} targetting column {}'.format(i, target * 2)

            if target * 2 == i:
                # This column is already in the correct place
                assert self.new_mapping[i] == -1
                self.new_mapping[i] = target
            elif self.new_mapping[i - 1] < 0:
                # Nothing is to the left. Move to the left by one.
                self.new_mapping[i - 1] = target
                # If there isn't already an edge moving horizontally select this one.
                if horizontal_edge == -1:
                    horizontal_edge = i
                    horizontal_edge_target = target
                    # The variable target is the index of the graph column, and
                    # therefore target * 2 + 3 is the actual screen column of
                    # the first horizontal line.
                    for j in range((target * 2) + 3, i - 2, 2):
                        self.new_mapping[j] = target
            elif self.new_mapping[i - 1] == target:
                # There is a branch line to our left already, and it is our
                # target. We combine with this line, since we share the same
                # parent commit.
                #
                # We don't have to add anything to the output or new_mapping,
                # since the existing branch line has already taken care of it.
                pass
            else:
                # There is a branch line to our left, but it isn't our target.
                # We need to cross over it.
                #
                # The space just to the left of this branch should always be empty.
                #
                # The branch to the left of that space should be our eventual target.
                assert self.new_mapping[i - 1] > target
                assert self.new_mapping[i - 2] < 0
                assert self.new_mapping[i - 3] == target
                self.new_mapping[i - 2] = target
                # Mark this branch as the horizontal edge to prevent any other
                # edges from moving horizontally.
                if horizontal_edge == -1:
                    horizontal_edge = i

        # The new mapping may be 1 smaller than the old mapping
        if self.new_mapping[self.mapping_size - 1] < 0:
            self.mapping_size -= 1

        # Output a line based on the new mapping info
        for i in range(self.mapping_size):
            target = self.new_mapping[i]
            if target < 0:
                self.buf += ' '
            elif target * 2 == i:
                self._write_column(self.new_columns[target], '|')
            elif target == horizontal_edge_target and i != horizontal_edge - 1:
                # Set the mappings for all but the first segment to -1 so that
                # they won't continue into the next line.
                if i != (target * 2) + 3:
                    self.new_mapping[i] = -1
                used_horizontal = True
                self._write_column(self.new_columns[target], '_')
            else:
                if used_horizontal and i < horizontal_edge:
                    self.new_mapping[i] = -1
                self._write_column(self.new_columns[target], '/')

        self._pad_horizontally(self.mapping_size)
        self.mapping, self.new_mapping = self.new_mapping, self.mapping

        # If self.mapping indicates that all of the branch lines are already in
        # the correct positions, we are done. Otherwise, we need to collapse
        # some branch lines together.
        if self._is_mapping_correct():
            self._update_state(GraphState.PADDING)

    def _next_line(self):  # pylint: disable=too-many-return-statements
        if self.state == GraphState.PADDING:
            self._output_padding_line()
            return False
        elif self.state == GraphState.SKIP:
            self._output_skip_line()
            return False
        elif self.state == GraphState.PRE_COMMIT:
            self._output_pre_commit_line()
            return False
        elif self.state == GraphState.COMMIT:
            self._output_commit_line()
            return True
        elif self.state == GraphState.POST_MERGE:
            self._output_post_merge_line()
            return False
        elif self.state == GraphState.COLLAPSING:
            self._output_collapsing_line()
            return False
        else:
            return False

    def _padding_line(self):
        """Output a padding line in the graph.

        This is similar to next_line(). However, it is guaranteed to never print
        the current commit line. Instead, if the commit line is next, it will
        simply output a line of vertical padding, extending the branch lines
        downwards, but leaving them otherwise unchanged.
        """
        if self.state != GraphState.COMMIT:
            self._next_line()
            return

        # Output the row containing this commit
        # Iterate up to and including self.num_columns, since the current commit
        # may not be in any of the existing columns. (This happens when the
        # current commit doesn't have any children that we have already
        # processed.)
        for i in range(self.num_columns):
            col = self.columns[i]
            self._write_column(col, '|')
            if col.commit == self.commit and self.num_parents > 2:
                self.buf += ' ' * (self.num_parents - 2) * 2
            else:
                self.buf += ' '

        self._pad_horizontally(self.num_columns)

        # Update self.prev_state since we have output a padding line
        self.prev_state = GraphState.PADDING

    def _is_commit_finished(self):
        return self.state == GraphState.PADDING

    def _show_commit(self):
        shown_commit_line = False

        # When showing a diff of a merge against each of its parents, we are
        # called once for each parent without update having been called. In this
        # case, simply output a single padding line.
        if self._is_commit_finished():
            self._show_padding()
            shown_commit_line = True

        while not shown_commit_line and not self._is_commit_finished():
            shown_commit_line = self._next_line()
            self.outfile.write(self.buf)
            if not shown_commit_line:
                self.outfile.write('\n')
            self.buf = ''

    def _show_padding(self):
        self._padding_line()
        self.outfile.write(self.buf)
        self.buf = ''

    def _show_remainder(self):
        shown = False

        if self._is_commit_finished():
            return False

        while True:
            self._next_line()
            self.outfile.write(self.buf)
            self.buf = ''
            shown = True

            if not self._is_commit_finished():
                self.outfile.write('\n')
            else:
                break

        return shown



import numpy as np

from .base import ColumnBase
from .descriptors import OneOf, String, OptionalString, SizedIntegerTuple, OptionalDict
from ..records import CompatibleData


@OneOf(['variable_shape', 'fixed_shape'])
class NdarraySchemaType(String):
    pass


@OneOf(['ndarray'])
class NdarrayColumnType(String):
    pass


@OneOf(['0'])
class DataHasherTcode(String):
    pass


class NdarraySchemaBase(ColumnBase):
    _schema_type = NdarraySchemaType()
    _column_type = NdarrayColumnType()
    _data_hasher_tcode = DataHasherTcode()

    def __init__(
            self,
            shape,
            dtype,
            backend=None,
            backend_options=None,
            *args, **kwargs
    ):
        if 'data_hasher_tcode' not in kwargs:
            kwargs['data_hasher_tcode'] = '0'
        super().__init__(*args, **kwargs)

        if backend_options is not None and backend is None:
            raise ValueError(
                '`backend_options` cannot be set if `backend` is not also provided.')

        if not isinstance(dtype, str):
            dtype = np.dtype(dtype).name
        self._dtype = dtype
        self._shape = shape
        self._backend = backend
        self._backend_options = backend_options
        self._schema_attributes.extend(
            ['_schema_type', '_shape', '_dtype', '_backend', '_backend_options'])

    def backend_from_heuristics(self):
        # uncompressed numpy memmap data is most appropriate for data whose shape is
        # likely small tabular row data (CSV or such...)
        if (len(self._shape) == 1) and (self._shape[0] < 400):
            backend = '10'
        # hdf5 is the default backend for larger array sizes.
        elif (len(self._shape) == 1) and (self._shape[0] <= 10_000_000):
            backend = '00'
        # on fixed arrays sized arrays apply optimizations.
        elif self._schema_type == 'fixed_shape':
            backend = '01'
        else:
            backend = '00'
        self._backend = backend

    @property
    def schema_type(self):
        return self._schema_type

    @property
    def shape(self):
        return self._shape

    @property
    def dtype(self):
        return np.dtype(self._dtype)

    @property
    def backend(self):
        return self._backend

    @property
    def backend_options(self):
        return self._backend_options

    def data_hash_digest(self, data: np.ndarray) -> str:
        return self._data_hasher_func(data)

    def change_backend(self, backend, backend_options=None):
        old_backend = self._backend
        old_backend_options = self._backend_options
        try:
            del self._beopts
            self._backend = backend
            self._beopts = backend_options
            self._backend_options = self._beopts.backend_options
        except (TypeError, ValueError) as e:
            del self._beopts
            self._backend = old_backend
            self._beopts = old_backend_options
            self._backend_options = self._beopts.backend_options
            raise e from None


@OneOf(['00', '01', '10', '50', None])
class NdarrayFixedShapeBackends(OptionalString):
    pass


@OneOf(['fixed_shape'])
class FixedShapeSchemaType(String):
    pass


class NdarrayFixedShape(NdarraySchemaBase):
    _shape = SizedIntegerTuple(size=31)
    _dtype = String()
    _backend = NdarrayFixedShapeBackends()
    _backend_options = OptionalDict()
    _schema_type = FixedShapeSchemaType()

    def __init__(self, *args, **kwargs):
        if 'column_type' in kwargs:
            super().__init__(*args, **kwargs)
        else:
            super().__init__(column_type='ndarray', *args, **kwargs)

        if 'schema_type' in kwargs:
            self._schema_type = kwargs['schema_type']
        else:
            self._schema_type = 'fixed_shape'

        if self.backend is None:
            self.backend_from_heuristics()
        self._backend_options = self._beopts.backend_options

    def verify_data_compatible(self, data):
        compatible = True
        reason = ''

        if not isinstance(data, np.ndarray):
            compatible = False
            reason = f'`data` argument type: {type(data)} != `np.ndarray`'
        elif data.dtype != self._dtype:
            compatible = False
            reason = f'dtype: {data.dtype.name} != aset: {self._dtype}.'
        elif not data.flags.c_contiguous:
            compatible = False
            reason = f'`data` must be "C" contiguous array.'
        elif data.shape != self._shape:
            compatible = False
            reason = f'data shape {data.shape} != fixed schema {self._shape}'

        res = CompatibleData(compatible, reason)
        return res


@OneOf(['00', '10', '50', None])
class NdarrayVariableShapeBackends(OptionalString):
    pass


@OneOf(['variable_shape'])
class VariableShapeSchemaType(String):
    pass


class NdarrayVariableShape(NdarraySchemaBase):
    _shape = SizedIntegerTuple(size=31)
    _dtype = String()
    _backend = NdarrayVariableShapeBackends()
    _backend_options = OptionalDict()
    _schema_type = VariableShapeSchemaType()

    def __init__(self, *args, **kwargs):
        if 'column_type' in kwargs:
            super().__init__(*args, **kwargs)
        else:
            super().__init__(column_type='ndarray', *args, **kwargs)

        if 'schema_type' in kwargs:
            self._schema_type = kwargs['schema_type']
        else:
            self._schema_type = 'variable_shape'

        if self.backend is None:
            self.backend_from_heuristics()
        self._backend_options = self._beopts.backend_options

    def verify_data_compatible(self, data):
        compatible = True
        reason = ''

        if not isinstance(data, np.ndarray):
            compatible = False
            reason = f'`data` argument type: {type(data)} != `np.ndarray`'
        elif data.dtype != self._dtype:
            compatible = False
            reason = f'dtype: {data.dtype.name} != aset: {self._dtype}.'
        elif not data.flags.c_contiguous:
            compatible = False
            reason = f'`data` must be "C" contiguous array.'
        elif data.ndim != len(self._shape):
            compatible = False
            reason = f'data rank {data.ndim} != aset rank {len(self._shape)}'
        elif not all([(dim <= maxdim) for dim, maxdim in zip(data.shape, self._shape)]):
            compatible = False
            reason = f'shape {data.shape} exceeds schema max {self._shape}'

        res = CompatibleData(compatible, reason)
        return res



from .utils import is_64bits, parse_bytes

# parsing constants

SEP_KEY = ':'
SEP_LST = ' '
SEP_CMT = ' << '
SEP_SLC = "*"
SEP_HSH = '$'

CMT_KV_JOIN_KEY = SEP_LST.encode()
CMT_DIGEST_JOIN_KEY = ''
CMT_REC_JOIN_KEY = SEP_HSH.encode()

K_INT = f'#'  # must be length 1 value
K_BRANCH = f'branch{SEP_KEY}'
K_HEAD = 'head'
K_REMOTES = f'remote{SEP_KEY}'
K_STGARR = f'a{SEP_KEY}'
K_STGMETA = f'l{SEP_KEY}'
K_SCHEMA = f's{SEP_KEY}'
K_HASH = f'h{SEP_KEY}'
K_WLOCK = f'writerlock{SEP_KEY}'
K_VERSION = 'software_version'

WLOCK_SENTINAL = 'LOCK_AVAILABLE'

# directory names

DIR_HANGAR = '.hangar'
DIR_HANGAR_SERVER = '.hangar_server'
DIR_DATA = 'data'
DIR_DATA_STORE = 'store_data'
DIR_DATA_STAGE = 'stage_data'
DIR_DATA_REMOTE = 'remote_data'

# configuration file names:

CONFIG_USER_NAME = 'config_user.ini'
CONFIG_SERVER_NAME = 'config_server.ini'

# LMDB database names and settings.


LMDB_SETTINGS = {
    # lmdb cannot open map larger than 2GB on 32 bit machines.
    'map_size': 50_000_000_000,
    'meminit': False,
    'subdir': False,
    'lock': False,
    'max_spare_txns': 4,
}

LMDB_REF_NAME = 'ref.lmdb'
LMDB_HASH_NAME = 'hash.lmdb'
LMDB_BRANCH_NAME = 'branch.lmdb'
LMDB_STAGE_REF_NAME = 'stage_ref.lmdb'
LMDB_STAGE_HASH_NAME = 'stage_hash.lmdb'


# readme file

README_FILE_NAME = 'README.txt'
